# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'client.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainClientWindow(object):
    def setupUi(self, MainClientWindow):
        MainClientWindow.setObjectName("MainClientWindow")
        MainClientWindow.resize(730, 575)
        MainClientWindow.setMinimumSize(QtCore.QSize(730, 575))
        self.centralwidget = QtWidgets.QWidget(MainClientWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.label_contacts = QtWidgets.QLabel(self.centralwidget)
        self.label_contacts.setGeometry(QtCore.QRect(10, 7, 251, 16))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.label_contacts.setFont(font)
        self.label_contacts.setObjectName("label_contacts")
        self.label_history = QtWidgets.QLabel(self.centralwidget)
        self.label_history.setGeometry(QtCore.QRect(270, 7, 441, 16))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.label_history.setFont(font)
        self.label_history.setObjectName("label_history")
        self.label_new_message = QtWidgets.QLabel(self.centralwidget)
        self.label_new_message.setGeometry(QtCore.QRect(270, 377, 441, 16))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.label_new_message.setFont(font)
        self.label_new_message.setObjectName("label_new_message")
        self.list_contacts = QtWidgets.QListView(self.centralwidget)
        self.list_contacts.setGeometry(QtCore.QRect(10, 30, 250, 450))
        self.list_contacts.setObjectName("list_contacts")
        self.list_messages = QtWidgets.QListView(self.centralwidget)
        self.list_messages.setGeometry(QtCore.QRect(270, 30, 450, 335))
        self.list_messages.setObjectName("list_messages")
        self.text_message = QtWidgets.QTextEdit(self.centralwidget)
        self.text_message.setGeometry(QtCore.QRect(270, 400, 450, 80))
        self.text_message.viewport().setProperty(
            "cursor", QtGui.QCursor(QtCore.Qt.IBeamCursor))
        self.text_message.setObjectName("text_message")
        self.btn_add_contact = QtWidgets.QPushButton(self.centralwidget)
        self.btn_add_contact.setGeometry(QtCore.QRect(9, 487, 122, 36))
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setKerning(True)
        self.btn_add_contact.setFont(font)
        self.btn_add_contact.setCursor(
            QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        self.btn_add_contact.setObjectName("btn_add_contact")
        self.btn_remove_contact = QtWidgets.QPushButton(self.centralwidget)
        self.btn_remove_contact.setGeometry(QtCore.QRect(139, 487, 122, 36))
        font = QtGui.QFont()
        font.setPointSize(12)
        self.btn_remove_contact.setFont(font)
        self.btn_remove_contact.setCursor(
            QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        self.btn_remove_contact.setObjectName("btn_remove_contact")
        self.btn_clear = QtWidgets.QPushButton(self.centralwidget)
        self.btn_clear.setGeometry(QtCore.QRect(469, 487, 122, 36))
        font = QtGui.QFont()
        font.setPointSize(12)
        self.btn_clear.setFont(font)
        self.btn_clear.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        self.btn_clear.setObjectName("btn_clear")
        self.btn_send = QtWidgets.QPushButton(self.centralwidget)
        self.btn_send.setGeometry(QtCore.QRect(599, 487, 122, 36))
        font = QtGui.QFont()
        font.setPointSize(12)
        self.btn_send.setFont(font)
        self.btn_send.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        self.btn_send.setObjectName("btn_send")
        MainClientWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainClientWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 730, 21))
        self.menubar.setObjectName("menubar")
        self.menu = QtWidgets.QMenu(self.menubar)
        self.menu.setObjectName("menu")
        self.menu_2 = QtWidgets.QMenu(self.menubar)
        self.menu_2.setObjectName("menu_2")
        MainClientWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainClientWindow)
        self.statusbar.setObjectName("statusbar")
        MainClientWindow.setStatusBar(self.statusbar)
        self.menu_exit = QtWidgets.QAction(MainClientWindow)
        self.menu_exit.setObjectName("menu_exit")
        self.menu_add_contact = QtWidgets.QAction(MainClientWindow)
        self.menu_add_contact.setObjectName("menu_add_contact")
        self.menu_del_contact = QtWidgets.QAction(MainClientWindow)
        self.menu_del_contact.setObjectName("menu_del_contact")
        self.menu.addAction(self.menu_exit)
        self.menu_2.addAction(self.menu_add_contact)
        self.menu_2.addAction(self.menu_del_contact)
        self.menubar.addAction(self.menu.menuAction())
        self.menubar.addAction(self.menu_2.menuAction())

        self.retranslateUi(MainClientWindow)
        QtCore.QMetaObject.connectSlotsByName(MainClientWindow)

    def retranslateUi(self, MainClientWindow):
        _translate = QtCore.QCoreApplication.translate
        MainClientWindow.setWindowTitle(
            _translate("MainClientWindow", "Чат программа Alpha-релиз"))
        self.label_contacts.setText(
            _translate("MainClientWindow", "Список контактов:"))
        self.label_history.setText(
            _translate("MainClientWindow", "История сообщений:"))
        self.label_new_message.setText(
            _translate("MainClientWindow", "Введите новое сообщение:"))
        self.btn_add_contact.setText(
            _translate("MainClientWindow", "+ Добавить"))
        self.btn_remove_contact.setText(
            _translate("MainClientWindow", "- Удалить"))
        self.btn_clear.setText(_translate("MainClientWindow", "Очистить"))
        self.btn_send.setText(_translate("MainClientWindow", "Отправить"))
        self.menu.setTitle(_translate("MainClientWindow", "Файл"))
        self.menu_2.setTitle(_translate("MainClientWindow", "Контакты"))
        self.menu_exit.setText(_translate("MainClientWindow", "Выход"))
        self.menu_add_contact.setText(
            _translate("MainClientWindow", "Добавить контакт"))
        self.menu_del_contact.setText(
            _translate("MainClientWindow", "Удалить контакт"))



import os.path
from datetime import datetime

from sqlalchemy import create_engine, MetaData, Table, Column, Integer, \
    String, Text, DateTime
from sqlalchemy.orm import mapper, sessionmaker
from sqlalchemy.sql import default_comparator


class ClientDatabase:
    """
    Класс - оболочка для работы с базой данных клиента.
    Использует SQLite базу данных, реализован с помощью SQLAlchemy ORM
    и используется классический подход.
    """

    class KnownUsers:
        """Класс - отображение для таблицы всех пользователей."""

        def __init__(self, user):
            self.id = None
            self.username = user

    class MessageStat:
        """Класс - отображение для таблицы статистики переданных сообщений."""

        def __init__(self, contact, direction, message):
            self.id = None
            self.contact = contact
            self.direction = direction
            self.message = message
            self.date = datetime.now()

    class Contacts:
        """Класс - отображение для таблицы контактов."""

        def __init__(self, name):
            self.id = None
            self.name = name

    def __init__(self, name):
        path = os.getcwd()
        filename = f'client_{name}.db3'
        self.db_engine = create_engine(
            f'sqlite:///{os.path.join(path, filename)}',
            echo=False,
            pool_recycle=7200,
            connect_args={'check_same_thread': False})
        self.metadata = MetaData()

        # Известные пользователи
        users = Table('known_users', self.metadata,
                      Column('id', Integer, primary_key=True),
                      Column('username', String))

        # История сообщений
        history = Table('message_history', self.metadata,
                        Column('id', Integer, primary_key=True),
                        Column('contact', String),
                        Column('direction', String),
                        Column('message', Text),
                        Column('date', DateTime))

        # Контакты
        contacts = Table('contacts', self.metadata,
                         Column('id', Integer, primary_key=True),
                         Column('name', String, unique=True))

        self.metadata.create_all(self.db_engine)

        # Отображения
        mapper(self.KnownUsers, users)
        mapper(self.MessageStat, history)
        mapper(self.Contacts, contacts)

        Session = sessionmaker(bind=self.db_engine)
        self.session = Session()

        self.session.query(self.Contacts).delete()
        self.session.commit()

    def add_contact(self, contact):
        """Метод добавляющий контакт в базу данных."""
        if not self.session.query(
                self.Contacts).filter_by(name=contact).count():
            contact_row = self.Contacts(contact)
            self.session.add(contact_row)
            self.session.commit()

    def contacts_clear(self):
        """Метод очищающий таблицу со списком контактов."""
        self.session.query(self.Contacts).delete()

    def del_contact(self, contact):
        """Метод удаляющий определённый контакт."""
        self.session.query(self.Contacts).filter_by(name=contact).delete()

    def add_users(self, users_list):
        """Метод заполняющий таблицу известных пользователей."""
        self.session.query(self.KnownUsers).delete()

        for user in users_list:
            user_row = self.KnownUsers(user)
            self.session.add(user_row)
        self.session.commit()

    def save_message(self, contact, direction, message):
        """Метод сохраняющий сообщение в базе данных."""
        message_row = self.MessageStat(contact, direction, message)
        self.session.add(message_row)
        self.session.commit()

    def get_contacts(self):
        """Метод возвращающий список всех контактов."""
        return [contact[0] for contact in self.session.query(
            self.Contacts.name).all()]

    def get_users(self):
        """Метод возвращающий список всех известных пользователей."""
        return [user[0] for user in self.session.query(
            self.KnownUsers.username).all()]

    def check_user(self, user):
        """Метод проверяющий существует ли пользователь."""
        if self.session.query(
                self.KnownUsers).filter_by(username=user).count():
            return True
        return False

    def check_contact(self, contact):
        """Метод проверяющий существует ли контакт."""
        if self.session.query(self.Contacts).filter_by(name=contact).count():
            return True
        return False

    def get_history(self, contact):
        """
        Метод возвращающий историю сообщений с определённым пользователем.
        """
        query = self.session.query(self.MessageStat).filter_by(contact=contact)
        return [(hist_row.contact,
                 hist_row.direction,
                 hist_row.message,
                 hist_row.date) for hist_row in query.all()]


if __name__ == '__main__':
    test_db = ClientDatabase('wasya')
    # for contact in ['masya', 'beaver', 'ipsum']:
    #     test_db.add_contact(contact)
    # test_db.add_contact('someone')
    #
    # test_db.add_users(['test1', 'test2', 'test3', 'test4', 'test5'])
    # test_db.save_message('test1', 'test2', f'Привет! я тестовое сообщение, время: {datetime.now()}!')
    # test_db.save_message('test2', 'test1', f'Привет! я другое тестовое сообщение, время: {datetime.now()}!')
    #
    # print(test_db.get_contacts())
    # print(test_db.get_users())
    # print(test_db.check_user('test1'))
    # print(test_db.check_user('test10'))
    # print(test_db.get_history('test2'))
    # print(test_db.get_history('test3'))
    # test_db.del_contact('someone')
    # print(test_db.get_contacts())

    # print(sorted(test_db.get_history('test2'), key=lambda item: item[3]))



class ServerError(Exception):
    """
    Класс - исключение, для обработки ошибок сервера.
    При генерации требует строку с описанием ошибки, полученную с сервера.
    """

    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text


# Отсутствует обязательное поле в принятом словаре
# class ReqFieldMissingError(Exception):
#     def __init__(self, missing_field):
#         self.missing_field = missing_field
#
#     def __str__(self):
# return f'В принятом словаре отсутствует обязательное поле
# {self.missing_field}!'



from socket import socket
from sys import argv
from logging import getLogger

import logs.config_server_log
import logs.config_client_log

# метод определения модуля, источника запуска.
if argv[0].find('client') == -1:
    # если не клиент то сервер!
    logger = getLogger('server')
else:
    # ну, раз не сервер, то клиент
    logger = getLogger('client')


def log(func_to_log):
    """
    Декоратор, выполняющий логирование вызовов функций.
    Сохраняет события типа debug,
    содержащие информацию об имени вызываемой функиции,
    параметры с которыми вызывается функция, и модуль, вызывающий функцию.
    """
    def log_saver(*args, **kwargs):
        logger.debug(
            f'Была вызвана функция {func_to_log.__name__} '
            f'c параметрами {args}, {kwargs}. '
            f'Вызов из модуля {func_to_log.__module__}')
        ret = func_to_log(*args, **kwargs)
        return ret
    return log_saver


def login_required(func):
    """
    Декоратор, проверяющий, что клиент авторизован на сервере.
    Проверяет, что передаваемый объект сокета находится
    в списке авторизованных клиентов.
    За исключением передачи словаря-запроса на авторизацию.
    Если клиент не авторизован, генерирует исключение TypeError
    """

    def checker(*args, **kwargs):
        # проверяем, что первый аргумент - экземпляр MessageProcessor
        # Импортить необходимо тут, иначе ошибка рекурсивного импорта.
        from server.core import MessageProcessor
        from common.variables import ACTION, PRESENCE
        if isinstance(args[0], MessageProcessor):
            found = False
            for arg in args:
                if isinstance(arg, socket):
                    # Проверяем, что данный сокет есть в списке names класса
                    # MessageProcessor
                    for client in args[0].names:
                        if args[0].names[client] == arg:
                            found = True

            # Теперь надо проверить, что передаваемые аргументы
            # не presence сообщение. Если presence, то разрешаем
            for arg in args:
                if isinstance(arg, dict):
                    if ACTION in arg and arg[ACTION] == PRESENCE:
                        found = True
            # Если не авторизован и не сообщение начала авторизации,
            # то вызываем исключение.
            if not found:
                raise TypeError
        return func(*args, **kwargs)

    return checker



from binascii import hexlify, b2a_base64
from hashlib import pbkdf2_hmac
import hmac
from json import JSONDecodeError
from logging import getLogger
from socket import socket, AF_INET, SOCK_STREAM
from threading import Lock, Thread
from time import sleep, time

from PyQt5.QtCore import QObject, pyqtSignal

from common.errors import ServerError
from common.variables import ATTEMPTS, ACTION, PRESENCE, TIME, USER, \
    ACCOUNT_NAME, RESPONSE, ERROR, MESSAGE, SENDER, DESTINATION, \
    MESSAGE_TEXT, GET_CONTACTS, LIST_INFO, USERS_REQUEST, ADD_CONTACT, \
    REMOVE_CONTACT, EXIT, PUBLIC_KEY, DATA, RESPONSE_511, PUBLIC_KEY_REQUEST
from common.utils import send_message, get_message

# Инициализация клиентского логера
logger = getLogger('client')
socket_lock = Lock()


class ClientTransport(Thread, QObject):
    """
    Класс реализующий транспортную подсистему клиентского модуля.
    Отвечает за взаимодействие с сервером.
    """
    # Сигналы: новое сообщение и потеря соединения
    new_message = pyqtSignal(dict)
    message_205 = pyqtSignal()
    connection_lost = pyqtSignal()

    def __init__(self, port, ip_address, db, username, password, keys):
        Thread.__init__(self)
        QObject.__init__(self)

        self.db = db
        self.username = username
        self.password = password
        self.transport = None
        self.keys = keys
        self.connection_init(port, ip_address)

        try:
            self.user_list_update()
            self.contacts_list_update()
        except OSError as err:
            if err.errno:
                logger.critical('Потеряно соединение с сервером')
                raise ServerError('Потеряно соединение с сервером!')
            logger.error('Timeout при обновлении списков пользователей')
        except JSONDecodeError:
            logger.critical('Потеряно соединение с сервером')
            raise ServerError('Потеряно соединение с сервером!')
        self.running = True

    def connection_init(self, port, ip):
        """Метод отвечающий за устанновку соединения с сервером."""
        self.transport = socket(AF_INET, SOCK_STREAM)
        self.transport.settimeout(5)

        connected = False

        for attempt in range(ATTEMPTS):
            logger.info(f'Попытка подключения №{attempt + 1}')
            try:
                self.transport.connect((ip, port))
            except (OSError, ConnectionRefusedError):
                pass
            else:
                connected = True
                logger.debug("Соединение установлено")
                break
            sleep(1)

        if not connected:
            logger.critical('Не удалось установить соединение с сервером')
            raise ServerError('Не удалось установить соединение с сервером')
        logger.debug('Диалог запуска авторизации')

        # Запускаем процедуру авторизации
        # Получаем хэш пароля
        password_bytes = self.password.encode('utf-8')
        salt = self.username.lower().encode('utf-8')
        password_hash = pbkdf2_hmac('sha512', password_bytes, salt, 10000)
        password_hash_string = hexlify(password_hash)

        logger.debug(f'Хэш пароля готов: {password_hash_string}')

        # Получаем публичный ключ и декодируем его из байтов
        pubkey = self.keys.publickey().export_key().decode('ascii')

        # Авторизируемся на сервере
        with socket_lock:
            presence = {ACTION: PRESENCE,
                        TIME: time(),
                        USER: {
                            ACCOUNT_NAME: self.username,
                            PUBLIC_KEY: pubkey}}
            logger.debug(f'Сообщение присутствия = {presence}')

            try:
                send_message(self.transport, presence)
                ans = get_message(self.transport)
                logger.debug(f'Ответ сервера = {ans}.')

                if RESPONSE in ans:
                    if ans[RESPONSE] == 400:
                        raise ServerError(ans[ERROR])
                    elif ans[RESPONSE] == 511:
                        ans_data = ans[DATA]
                        hash_ = hmac.new(password_hash_string,
                                        ans_data.encode('utf-8'), 'MD5')
                        digest = hash_.digest()
                        ans_ = RESPONSE_511
                        ans_[DATA] = b2a_base64(digest).decode('ascii')
                        send_message(self.transport, ans_)
                        self.process_server_ans(get_message(self.transport))
            except (OSError, JSONDecodeError) as err:
                logger.debug('Ошибка связи', exc_info=err)
                raise ServerError('Сбой соединения в процессе авторизации')

    def process_server_ans(self, message):
        """Метод обработчик поступающих сообщений с сервера."""
        logger.debug(f'Разбор сообщения от сервера: {message}')

        if RESPONSE in message:
            if message[RESPONSE] == 200:
                return
            elif message[RESPONSE] == 400:
                raise ServerError(f'{message[ERROR]}')
            elif message[RESPONSE] == 205:
                self.user_list_update()
                self.contacts_list_update()
                self.message_205.emit()
            else:
                logger.debug(f'Принят неизвестный код подтверждения '
                             f'{message[RESPONSE]}')

        elif ACTION in message and message[ACTION] == MESSAGE \
                and SENDER in message and DESTINATION in message \
                and MESSAGE_TEXT in message \
                and message[DESTINATION] == self.username:
            logger.debug(f'Получено сообщение от пользователя '
                         f'{message[SENDER]}: {message[MESSAGE_TEXT]}')
            self.new_message.emit(message)

    def contacts_list_update(self):
        """Метод обновляющий с сервера список контактов."""
        logger.debug(f'Запрос контактов для пользователя {self.name}')
        req = {
            ACTION: GET_CONTACTS,
            TIME: time(),
            USER: self.username
        }
        logger.debug(f'Сформирован запрос {req}')

        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        logger.debug(f'Получен ответ {ans}')
        if RESPONSE in ans and ans[RESPONSE] == 202:
            for contact in ans[LIST_INFO]:
                self.db.add_contact(contact)
        else:
            logger.error('Не удалось обновить список контактов')

    def user_list_update(self):
        """Метод обновляющий с сервера список пользователей."""
        logger.debug(f'Запрос списка известных пользователей {self.username}')
        req = {
            ACTION: USERS_REQUEST,
            TIME: time(),
            ACCOUNT_NAME: self.username
        }

        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        if RESPONSE in ans and ans[RESPONSE] == 202:
            self.db.add_users(ans[LIST_INFO])
        else:
            logger.error('Не удалось обновить список известных пользователей')

    def key_request(self, user):
        """Метод запрашивающий с сервера публичный ключ пользователя."""
        logger.debug(f'Запрос публичного ключа для {user}')
        req = {
            ACTION: PUBLIC_KEY_REQUEST,
            TIME: time(),
            ACCOUNT_NAME: user
        }
        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        if RESPONSE in ans and ans[RESPONSE] == 511:
            return ans[DATA]
        logger.error(f'Не удалось получить ключ собеседника {user}.')

    # Сообщает на сервер о добавлении нового контакта
    def add_contact(self, contact):
        """Метод отправляющий на сервер сведения о добавлении контакта."""
        logger.debug(f'Создание контакта {contact}')
        req = {
            ACTION: ADD_CONTACT,
            TIME: time(),
            USER: self.username,
            ACCOUNT_NAME: contact
        }

        with socket_lock:
            send_message(self.transport, req)
            self.process_server_ans(get_message(self.transport))

    def remove_contact(self, contact):
        """Метод отправляющий на сервер сведения о удалении контакта."""
        logger.debug(f'Удаление контакта {contact}')
        req = {
            ACTION: REMOVE_CONTACT,
            TIME: time(),
            USER: self.username,
            ACCOUNT_NAME: contact
        }

        with socket_lock:
            send_message(self.transport, req)
            self.process_server_ans(get_message(self.transport))

    def transport_shutdown(self):
        """Метод уведомляющий сервер о завершении работы клиента."""
        self.running = False
        message = {
            ACTION: EXIT,
            TIME: time(),
            ACCOUNT_NAME: self.username
        }

        with socket_lock:
            try:
                send_message(self.transport, message)
            except OSError:
                pass
        logger.debug('Транспорт завершает работу')
        sleep(0.5)

    def send_message(self, to, message):
        """Метод отправляющий на сервер сообщения для пользователя."""
        message_dict = {
            ACTION: MESSAGE,
            SENDER: self.username,
            DESTINATION: to,
            TIME: time(),
            MESSAGE_TEXT: message
        }
        logger.debug(f'Сформирован словарь сообщения: {message_dict}')

        # Дожидаемся освобождения сокета для отправки сообщения
        with socket_lock:
            send_message(self.transport, message_dict)
            self.process_server_ans(get_message(self.transport))
            logger.info(f'Отправлено сообщение для польователя {to}')

    def run(self):
        """Метод содержащий основной цикл работы транспортного потока."""
        logger.debug('Запущен процесс - приёмник сообщений от сервера')

        while self.running:
            sleep(1)
            message = None

            with socket_lock:
                try:
                    self.transport.settimeout(0.5)
                    message = get_message(self.transport)
                except OSError as err:
                    if err.errno:
                        logger.critical('Потеряно соединение с сервером')
                        self.running = False
                        self.connection_lost.emit()
                except (ConnectionError,
                        ConnectionAbortedError,
                        ConnectionResetError,
                        JSONDecodeError,
                        TypeError):
                    logger.debug('Потеряно соединение с сервером')
                    self.running = False
                    self.connection_lost.emit()
                finally:
                    self.transport.settimeout(5)

            # Если сообщение получено, то вызываем функцию обработчик
            if message:
                logger.debug(f'Принято сообщение с сервера: {message}')
                self.process_server_ans(message)



from dis import get_instructions


class ClientVerifier(type):
    """
    Метакласс, проверяющий что в результирующем классе нет серверных вызовов
    таких как: accept, listen.
    Также проверяется, что сокет не создаётся внутри конструктора класса.
    """

    def __init__(cls, clsname, bases, clsdict):
        methods = []

        for func in clsdict:
            try:
                instructions = get_instructions(clsdict[func])
            except TypeError:
                pass
            else:
                for instr in instructions:
                    if instr.opname == 'LOAD_GLOBAL' \
                            and instr.argval not in methods:
                        methods.append(instr.argval)

        for method in ('accept', 'listen', 'socket'):
            if method in methods:
                raise TypeError(f'В классе вызван запрещённый метод {method}')

        # Долго не мог определить почему не работает, спасибо  @log ;-)
        if 'get_message' in methods or 'send_message' in methods:
            pass
        else:
            raise TypeError(
                'Отсутствуют вызовы функций, работающих с сокетами')
        super().__init__(clsname, bases, clsdict)


class ServerVerifier(type):
    """
    Метакласс, проверяющий что в результирующем классе нет клиентских вызовов
    таких как: connect.
    Также проверяется, что серверный сокет является TCP
    и работает по IPv4 протоколу.
    """

    def __init__(cls, clsname, bases, clsdict):
        methods = []
        attrs = []

        for func in clsdict:
            try:
                instructions = get_instructions(clsdict[func])
            except TypeError:
                pass
            else:
                for instruction in instructions:
                    if instruction.opname == 'LOAD_GLOBAL' \
                            and instruction.argval not in methods:
                        methods.append(instruction.argval)
                    if instruction.opname == 'LOAD_ATTR' \
                            and instruction.argval not in attrs:
                        attrs.append(instruction.argval)

        if 'connect' in methods:
            raise TypeError(
                'Использование метода connect недопустимо в серверном классе')
        if not ('SOCK_STREAM' in attrs and 'AF_INET' in attrs):
            raise TypeError('Некорректная инициализация сокета')

        super().__init__(clsname, bases, clsdict)



from ipaddress import ip_address
from logging import getLogger
import sys


if sys.argv[0].find('client') == -1:
    logger = getLogger('server')
else:
    logger = getLogger('client')


class Port:
    """
    Класс - дескриптор для номера порта.
    Позволяет использовать только порты с 1023 по 65536.
    При попытке установить неподходящий номер порта генерирует исключение.
    """

    def __set__(self, instance, value):
        if not 1023 < value < 65536:
            logger.critical(
                f'Попытка запуска сервера с указанием неподходящего порта:'
                f'{value}. Допустимы адреса с 1024 до 65535.')
            raise TypeError('Некорректрый номер порта')
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name


class Addr:
    """
    Класс - дескриптор для номера порта.
    При попытке установить неподходящий IP-адрес генерирует исключение.
    """

    def __set__(self, instance, value):
        if value:
            try:
                ip = ip_address(value)
            except ValueError as e:
                logger.critical(f'Неверный ip адрес: {e}')
                sys.exit(1)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name



import os.path
import sys
from os import urandom
from argparse import ArgumentParser
from logging import getLogger

from Cryptodome.PublicKey import RSA
from PyQt5.QtWidgets import QApplication, QMessageBox

import logs.config_client_log
from common.decos import log
from common.variables import DEFAULT_IP_ADDRESS, DEFAULT_PORT
from common.errors import ServerError
from client.database import ClientDatabase
from client.transport import ClientTransport
from client.main_window import ClientMainWindow
from client.start_dialog import UserNameDialog

# Инициализация клиентского логера
logger = getLogger('client')


# Парсер аргументов коммандной строки
@log
def arg_parser():
    """
    Парсер аргументов командной строки, возвращает кортеж из 4 элементов:
    адрес сервера, порт, имя пользователя, пароль.
    Выполняет проверку на корректность номера порта.
    """
    parser = ArgumentParser()
    parser.add_argument('addr', default=DEFAULT_IP_ADDRESS, nargs='?')
    parser.add_argument('port', default=DEFAULT_PORT, type=int, nargs='?')
    parser.add_argument('-n', '--name', default=None, nargs='?')
    parser.add_argument('-p', '--password', default='', nargs='?')
    namespace = parser.parse_args(sys.argv[1:])
    server_address = namespace.addr
    server_port = namespace.port
    client_name = namespace.name
    client_password = namespace.password

    # проверим подходящий номер порта
    if not 1023 < server_port < 65536:
        logger.critical(
            f'Попытка запуска клиента с неподходящим номером порта: '
            f'{server_port}. Допустимы адреса с 1024 до 65535. '
            f'Клиент завершается.')
        sys.exit(1)

    return server_address, server_port, client_name, client_password


if __name__ == '__main__':
    # Загружаем параметы коммандной строки
    server_address, server_port, client_name, client_password = arg_parser()
    logger.debug('Аргументы загружены')

    client_app = QApplication(sys.argv)

    start_dialog = UserNameDialog()
    # Если имя пользователя не было задано, то запросим его
    if not client_name or not client_password:
        client_app.exec_()
        # Если пользователь ввёл имя и нажал OK,
        # то сохраняем введённое имя и удаляем объект или выходим
        if start_dialog.ok_pressed:
            client_name = start_dialog.client_name.text()
            client_password = start_dialog.client_password.text()
            logger.debug(f'Используются: ПОЛЬЗОВАТЕЛЬ = {client_name}, '
                         f'ПАРОЛЬ = {client_password}.')
        else:
            sys.exit(0)

    logger.info(
        f'Запущен клиент с парамертами: адрес сервера: {server_address}, '
        f'порт: {server_port}, имя пользователя: {client_name}')

    dir_path = os.getcwd()
    key_file = os.path.join(dir_path, f'{client_name}.key')

    if not os.path.exists(key_file):
        keys = RSA.generate(2048, urandom)
        with open(key_file, 'wb') as key:
            key.write(keys.export_key())
    else:
        with open(key_file, 'rb') as key:
            keys = RSA.import_key(key.read())

    logger.debug('Ключи успешно загружены')

    db = ClientDatabase(client_name)

    try:
        transport = ClientTransport(server_port, server_address, db,
                                    client_name, client_password, keys)
        logger.debug("Транспорт готов")
    except ServerError as err:
        message = QMessageBox()
        message.critical(start_dialog, 'Ошибка сервера', err.text)
        sys.exit(1)
    transport.setDaemon(True)
    transport.start()

    # Удалим объект диалога за ненадобностью
    del start_dialog

    # GUI
    main_window = ClientMainWindow(db, transport, keys)
    main_window.make_connection(transport)
    main_window.setWindowTitle(f'Чат программа Alpha-релиз - {client_name}')
    client_app.exec_()

    transport.transport_shutdown()
    transport.join()



from setuptools import setup, find_packages

setup(name="mrz_mess_client",
      version="0.0.1",
      description="mess_client",
      author="morozov1982",
      author_email="morozov1982@gmail.com",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex']
      )
